# -*- coding: utf-8 -*-
import pandas as pd
import re
import ipywidgets as widgets
from IPython.display import clear_output


from citableclass.base import Citable
from .base import CorpusTextSearch
# from corpussearch.base import CorpusTextSearch


class SearchWordGUI(widgets.HBox):
    """
    Helper Class, defines single logical element of according, to enable
    more complex search patterns.
    """
    def __init__(self, optionList, colName):
        super(SearchWordGUI, self).__init__()
        self.optList = optionList
        self.column = colName
        if self.column not in self.optList:
            self.optList.extend([colName])

        self._text = widgets.Text(description='Search for:', placeholder='term')
        self._text.layout.margin = '0 0 0 20px'

        self._select = widgets.Dropdown(
            description='Search in:',
            options=self.optList,
            value=self.column,
        )

        children = [self._text, self._select]
        self.children = children

    @property
    def value(self):
        return self._text.value

    @property
    def part(self):
        return self._select.value


class CorpusGUI(CorpusTextSearch):
    """
    GUI wrapper for 'CorpusTextSearch'-class, to be used in Jupyter Notebooks.
    Initialized with path to data. Search can be in main column 'colname' or
    other columns containing data in string format.

    Searches can be chained by using the 'more' button. Logical operations are
    AND, OR, NOT.

    Results are printed in the 'Sentence' field for the 'colname' cells,
    with corresponding metadata from the other columns in the 'Result' field.

    Navigation through results is possible using 'previous' and 'next' buttons.

    Example:
            >>> gui1 = CorpusGUI(
                            pathDF='/path/to/dataframe/file.xlsx',
                            dataType='excel',
                            dataIndex='single',
                            colname='main'
                            )

            >>> gui1.displayGUI()
            True
    """
    def __init__(
            self, pathDF,
            dataType='pickle', dataIndex='multi', colname='text',
            maxValues=2500, pathMeta=False, pathType=False
            ):

        super(CorpusGUI, self).__init__(pathDF, dataType, dataIndex, colname, maxValues, pathMeta, pathType)
        self.initSearch = SearchWordGUI(self.colValueDictTrigger, self.column)
        self.accordion = widgets.Accordion(children=[self.initSearch])
        self.accordion.set_title(0, 'Enter search term')

        self.extendSearch = widgets.ToggleButtons(options=['more', 'less'])
        self.extendSearch.on_msg(self._addSearchField)

        self.searchButton = widgets.Button(
            description='Search',
            button_style='danger'
        )

        self.searchButton.on_msg(self._searchLogic)

        self.direction = widgets.ToggleButtons(options=['previous', 'next'])

        self.direction.on_msg(self._setSentence)

        self.outInfo = widgets.Output()

        self.outSentence = widgets.Textarea(
            value='',
            layout=widgets.Layout(flex='0 1 auto', height='200px', min_height='40px', width='70%'),
            placeholder='Sentence'
        )

        self.outMeta = widgets.Textarea(
            placeholder='Result',
            layout=widgets.Layout(flex='0 1 auto', height='200px', min_height='40px', width='30%'),
            value=''
        )

    def _setDescription(self):
        """ Helper function to display metadata for result"""
        res = 'Result {0}\n'.format(self.counter) + '\n'.join(
             ["{0}:\n {1}".format(x, y) for x, y in self.displayResult.iloc[self.counter].to_dict().items() if x != self.column]
        )
        return res

    def _setSentence(self, widget, content, buffers):
        """ Helper function: Navigate and set result to display"""
        if widget.value == 'previous':
            self.counter = self.counter - 1
        elif widget.value == 'next':
            self.counter = self.counter + 1
        if self.counter < self.displayResult.shape[0] and self.counter > -1:
            self.outSentence.value = self.displayResult[self.column].iloc[self.counter]
            self.outMeta.value = self._setDescription()
        else:
            self.outSentence.value = 'End of found results. Enter new search.'
        return

    def _addSearchField(self, widget, content, buffers):
        """ Helper function to extend search logic."""
        child = SearchWordGUI(self.colValueDictTrigger, self.column)
        children = []
        for ch in self.accordion.children:
            children.append(ch)
        if self.extendSearch.value == 'more':
            children.append(
                widgets.Dropdown(
                    description='Select logic to connect words',
                    options={'and': '&', 'or': '|', 'not': '~&'},
                    value='&'
                )
            )
            children.append(child)
            self.accordion.children = children
            self.accordion.set_title(len(children)-2, 'and/or/not')
            self.accordion.set_title(len(children)-1, 'add term')
        elif self.extendSearch.value == 'less':
            if len(children) >= 3:
                lesschildren = children[:-2]
                self.accordion.children = lesschildren
            else:
                pass
            pass
        return

    def _searchLogic(self, widget, content, buffers):
        """
        Helper function to initialize logical reduction
        in CorpusTextSearch.logicReduce()
        """
        self.tempRes = []
        self.tempList = []
        self.counter = 0
        for ch in self.accordion.children:
            if ch.value not in ['&', '|', '~&']:
                self.tempList.append((ch.part, ch.value))
            else:
                self.tempList.append(ch.value)
        self.outputResult = self.logicReduce(self.tempList)

        self.displayResult = self.outputResult.results()

        if self.displayResult.shape[0] > 0:
            # TODO: Add plotting logic in handle_submit(sender)
            # handle_submit(sender)

            self.outSentence.value = re.sub(r'\n|\s+', ' ', self.displayResult[self.column].iloc[self.counter])
            self.outMeta.value = self._setDescription()
            with self.outInfo:
                clear_output()
                print('Found {} entries.'.format(self.displayResult.shape[0]))
        else:
            with self.outInfo:
                clear_output()
                print('Found no entries. Try changing search!')

    def displayGUI(self):
        """Display the GUI for CorpusTextSearch"""
        searchControl = widgets.HBox([self.extendSearch, self.searchButton, self.outInfo])
        textControl = widgets.HBox([self.direction])
        sentenceFields = widgets.HBox([self.outMeta, self.outSentence])
        searchBox = widgets.VBox([self.accordion, searchControl, textControl, sentenceFields])

        return display(searchBox)



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from collections import Counter, OrderedDict
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import ipywidgets as widgets
from IPython.display import clear_output


class CorpusVisualization(object):
    """
    Visualization layer for corpus search.
    """

    def __init__(self,search):
        self.search = search
        self.plotDict = {}
        plt.style.use('seaborn-deep')

        self.resetPlot = widgets.Button(
            description='Reset plot'
            )

        self.plotLevel = widgets.Text(
            description='Plot level',
            value=''
            )

        self.resetPlot.on_msg(
            self._resetPlotFunc(widget, content, buffers, self.plotLevel.value)
        )

        self.plotout = widgets.Output()

    def _resetPlotFunc(self, widget, content, buffers, level):
        """Clear all previous plots."""
        self.plotDict.clear()
        with self.plotout:
            self._initFigure(level)
            plt.show()

    def _initFigure(self,level):
        """Basic initalization for matplotlib figure."""
        clear_output()
        xticks = []
        try:
            if level in self.search.colValueDictTrigger and level != self.search.column:
                if level not in self.search.levelValues.keys():
                    if self.search.dataindex == 'multi':
                        self.search.levelValues[level] = self.search.dataframe.index.get_level_values(level).unique()
                    elif self.search.dataindex == 'single':
                        self.search.levelValues[level] = self.search.dataframe[level].unique()
                for vol in set(self.search.levelValues[level]):
                    xticks.append((vol))
        except ValueError:
            print('Can not use {0} for plotting'.format(level))

        xtickslabels = sorted(xticks, key=lambda x: x[0])

        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.set_xticks(np.linspace(1,  len(xtickslabels) + 1, len(xtickslabels), endpoint=False))
        ax.set_xticklabels(xtickslabels, rotation=45)
        ax.set_yticks(np.linspace(0.2, 1, 5))
        ax.set_yticklabels(['0', '', '0.5', '', '1'])

    def _countExp(self, expr, row):
        """Count occurance per level/column"""
        # TODO: redo this one
        pass

    def handle_submit(self, widget, content, buffers):
        resDict = {}
        iterate = [x[0] for x in self.result().iterrows()]
        resDict = OrderedDict(Counter((elem[0], elem[1]) for elem in iterate))
        with plotout:
            self._initFigure()
            labelStr = ' '.join([ch.value for ch in accordion.children])
            x = [x for x in range(1, len(resDict.keys())+1)]
            maxRadius = max(list(resDict.values()))
            radii = [xo for xo in [x/maxRadius for x in resDict.values()]]
            plotDict[labelStr] = (resDict.values(), maxRadius)
            i = 0
            maxVal = sorted([v[1] for key, v in plotDict.items()])[-1]
            colors = [cm.jet(i) for i in np.linspace(0, 1, len(plotDict.keys()))]
            for key, value in plotDict.items():
                radii = [xo for xo in [x/maxVal for x in value[0]]]
                ax.bar(x, radii, width=0.3, alpha=0.3, color=colors[i], label=key)
                i += 1

            ax.legend()

            plt.show()

    def showVisualButton(self):
        box = widgets.HBox([plotLevel,resetPlot])
        display(box)



# -*- coding: utf-8 -*-
import os
from ..base import CorpusTextSearch
from ..ml import CorpusML
from ..gui import CorpusGUI
from ..vis import CorpusVisualization

from unittest import TestCase

DATA = os.path.join(os.path.dirname(__file__), 'data')


class TestCorpusTextSearch(TestCase):
    """Basic import test"""
    def setUp(self):
        self.test1 = CorpusTextSearch(
            pathDF=os.path.join(DATA, 'dfTest.csv'),
            dataType='csv',
            dataIndex='single'
            )
        self.test2 = CorpusTextSearch(
            pathDF=os.path.join(DATA, 'dfTest.pickle'),
            dataType='pickle'
            )

    def TestImportException(self):
        self.assertRaises(TypeError, CorpusTextSearch)

    def TestImport(self):
        expected_datatype_1 = 'csv'
        expected_datatype_2 = 'pickle'
        self.assertEqual(self.test1.datatype, expected_datatype_1)
        self.assertEqual(self.test2.datatype, expected_datatype_2)

    def TestReduceSingleIndex(self):
        testdt = self.test1.reduce('part', 'Einleitung').reduce('paragraph', '1') \
                    .reduce('page', 10).results()
        self.assertEqual(testdt.shape[0], 1)
        self.assertEqual(testdt.volume.values[0], 'Band_1')

    def TestReduceSingleIndexExtResults(self):
        testdt = self.test2.reduce('part', 'Einleitung').reduce('paragraph', '1') \
                    .reduce('page', 10).extResults('part')
        self.assertEqual(testdt.shape[0], 1)
        self.assertEqual(testdt.words.values[0], 117)

    def TestReduceMultiIndex(self):
        testdt = self.test2.reduce('part', 'Einleitung').reduce('paragraph', '1') \
                    .reduce('page', 10).results()
        self.assertEqual(testdt.shape[0], 1)
        self.assertEqual(testdt.volume.values[0], 'Band_1')

    def TestReduceMultiIndexExtResults(self):
        testdt = self.test2.reduce('part', 'Einleitung').reduce('paragraph', '1') \
                    .reduce('page', 10).extResults('part')
        self.assertEqual(testdt.shape[0], 1)
        self.assertEqual(testdt.words.values[0], 117)


class TestCorpusML(TestCase):
    """Basic import test"""
    def setUp(self):
        self.test1 = CorpusML(
            pathDF=os.path.join(DATA, 'dfTest.csv'),
            dataType='csv',
            dataIndex='single'
            )
        self.test2 = CorpusML(
            pathDF=os.path.join(DATA, 'dfTest.pickle'),
            dataType='pickle'
            )

    def TestImportException(self):
        self.assertRaises(TypeError, CorpusML)


class TestCorpusGUI(TestCase):
    """Basic import test"""
    def setUp(self):
        self.test1 = CorpusGUI(
            pathDF=os.path.join(DATA, 'dfTest.csv'),
            dataType='csv',
            dataIndex='single'
            )
        self.test2 = CorpusGUI(
            pathDF=os.path.join(DATA, 'dfTest.pickle'),
            dataType='pickle'
            )

    def TestImportException(self):
        self.assertRaises(TypeError, CorpusGUI)


class TestCorpusVisual(TestCase):
    """Basic import test"""
    def setUp(self):
        self.test1 = CorpusVisualization()

    def TestImportException(self):
        def func1():
            try:
                CorpusVisualization(pathDF=os.path.join(DATA, 'dfTest.pickle'))
            except TypeError:
                raise
        self.assertRaises(TypeError, func1)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import pandas as pd
import re
import difflib
import logging
import os
import tempfile
import gensim
import requests
from citableclass.base import Citable
from .base import CorpusTextSearch

class CorpusML(CorpusTextSearch):
    """
    Machine learning routines for corpussearch.

    Trains gensim's Word2Vec for data contained in 'colname' column. Usefull
    to obtain word-contexts for additional searches. Expects additional
    inital key: language, to determine tokenizers, lemmatizers, and stopwords.
    Model parameters are given as a tuple of numbers,e.g.,
    '(workers=4,min_count=5,features=300)'.

    Currently supported languages (using CLTK and NLTK): latin, greek, english

    Example:
            >>> ml1 = CorpusML(
                            pathDF='/path/to/dataframe/file.xlsx',
                            dataType='excel',
                            dataIndex='single',
                            colname='main',
                            language='english'
                            )

            >>> ml1.trainModel()

            >>> ml1.getSimilarContext('word1')
            True
    """
    def __init__(
            self, pathDF, language='english',
            dataType='pickle', dataIndex='multi', colname='text',
            maxValues=2500, pathMeta=False, pathType=False, showLogging=False,
            model_params=(4,5,300)
            ):

        super(CorpusML, self).__init__(
            pathDF, dataType, dataIndex, colname,
            maxValues, pathMeta, pathType
            )

        if showLogging:
            logging.basicConfig(
                format='%(asctime)s : %(levelname)s : %(message)s',
                level=logging.INFO
                )

        self.model = gensim.models.Word2Vec(
            workers=model_params[0],
            min_count=model_params[1],
            size=model_params[2]
            )

        # self.model.random.seed(42)

        self.language = language

        if self.language == 'latin' or self.language == 'greek':
            from cltk.corpus.utils.importer import CorpusImporter
            corpus_importer = CorpusImporter(self.language)
            corpus_importer.import_corpus(
                '{0}_models_cltk'.format(self.language)
                )
            from cltk.stem.lemma import LemmaReplacer
            from cltk.tokenize.word import nltk_tokenize_words as tokenizer
            lemmatizer = LemmaReplacer(self.language)
            if self.language == 'latin':
                from cltk.stem.latin.j_v import JVReplacer
                from cltk.stop.latin.stops import STOPS_LIST as stopwords
                self.jvreplacer = JVReplacer()
            elif self.language == 'greek':
                from cltk.stop.greek.stops import STOPS_LIST as stopwords
        elif self.language == 'english' or 'german':
            import nltk
            nltk.download('stopwords')
            from nltk.stem import WordNetLemmatizer
            from nltk.tokenize import word_tokenize as tokenizer
            from nltk.corpus import stopwords
            stopwords = stopwords.words(self.language)
            lemmatizer = WordNetLemmatizer()
        else:
            raise ValueError(
                'Could not find lemmatizer, tokenizer,\
                 and stopwords for chosen language.')
        self.lemmatizer = lemmatizer
        self.tokenizer = tokenizer
        self.stopwords = stopwords

    def convert(self, row):
        """
        Helper function: Converts input data from 'colname'-column to a
        tokenized, lemmatized, and cleaned word list without stopwords.

        Additionaly, for latin language CLTK's jvreplace is applied.
        """
        reg = re.compile('[^a-zA-Z0-9]')
        if type(row) == str:
            sentence = self.tokenizer(row)
            sentence = [x.lower() for x in sentence]
            sentence = [x for x in sentence if x not in self.stopwords]
            sentence = [reg.sub('', x) for x in sentence]
            sentence = [x for x in sentence if x and len(x) > 1]
            sentence = [x.strip('0123456789') for x in sentence]
            if self.language == 'latin':
                sentence = [self.jvreplacer.replace(word) for word in sentence]
            try:
                sentence = self.lemmatizer.lemmatize(sentence)
            except:
                pass
            return sentence

    def saveTrainData(self):
        """
        Creates new training data column,
        by applying convert() to 'colname' column.
        """
        self.dataframe['training_data'] = self.dataframe[self.column].apply(
            lambda row: self.convert(row)
        )
        self.tempFile = tempfile.NamedTemporaryFile(delete=False)
        self.dataframe.to_pickle(self.tempFile.name)
        return

    def buildVocab(self):
        """
        Builds model vocabulary.
        """
        try:
            self.tempFile
        except AttributeError:
            self.saveTrainData()
        loadedData = pd.read_pickle(self.tempFile)
        self.training_data = loadedData.training_data.values.tolist()
        self.model.build_vocab(self.training_data)
        return

    def trainModel(self):
        """
        Trains model based on created training data.
        """
        try:
            self.training_data
        except AttributeError:
            self.buildVocab()
        self.model.train(
            self.training_data,
            total_examples=self.model.corpus_count,
            epochs=self.model.epochs
            )
        return

    def getSimilarContext(self, word):
        """
        Returns words occuring in the corpus with similar context. If the word
        is not contained in the training vocabulary, a similiar word is choosen
        using difflib's get_close_matches.
        """
        vocabulary = self.model.wv.vocab.keys()
        if word in vocabulary:
            res = self.model.wv.most_similar(word)
            return res
        else:
            simWord = difflib.get_close_matches(word, vocabulary, 1)
            print('Using similar word: {0}'.format(simWord[0]))
            res = self.model.wv.most_similar(simWord)
            return res



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import re
import difflib
from citableclass.base import Citable


class CorpusTextSearch(object):
    """
    This class is initialized with a searchstring, the name of the column,
    which contains the textstrings, and a path to a pickled dataframe.
    Optionally, a path to an excel file containing metadata can be provided.

    For other import formats change read_pickle and read_excel accordingly.

    Example:
            >>> search1 = CorpusTextSearch(
                            pathDF='/path/to/dataframe/file.xlsx',
                            dataType='excel',
                            dataIndex='single',
                            colname='main'
                            )

            >>> search1.reduce('date','1500').search('Thomas Morus').results()
            True
    """

    def __init__(
            self, pathDF,
            dataType='pickle', dataIndex='multi', colname='text',
            maxValues=2500, pathMeta=False, pathType=False
            ):

        self.datatype = dataType

        if pathType == 'DOI':
            pathDF = "http://dx.doi.org/{0}".format(pathDF)
        elif pathType == 'citable':
            self.citable = Citable(pathDF)
            self.dataframe = self.citable.digitalresource()

        if not pathType == 'citable':
            if self.datatype == 'pickle':
                self.dataframe = pd.read_pickle(pathDF)
            elif self.datatype == 'excel':
                self.dataframe = pd.read_excel(pathDF)
            elif self.datatype == 'csv':
                self.dataframe = pd.read_csv(pathDF, error_bad_lines=False)
            elif self.datatype == 'json':
                self.dataframe = pd.read_json(pathDF)
            else:
                raise ValueError(
                    'Please provide data in pickle,excel,csv or json format'
                    )

        self.dataindex = dataIndex

        self.colValueDictTrigger = []
        if self.dataindex == 'single':
            for col in self.dataframe.columns:
                if not self.dataframe[col].dtype.name in ['int', 'int64', 'float64']:
                    try:
                        length = len(self.dataframe[col].unique())
                        if length < maxValues:
                            self.colValueDictTrigger.append(col)
                    except TypeError:
                        print('Encountered list value in cell, skipping...')
                        pass
        elif self.dataindex == 'multi':
            for level in self.dataframe.index.names:
                if not self.dataframe.index.get_level_values(level).dtype.name in ['int', 'int64', 'float64']:
                    if len(self.dataframe.index.get_level_values(level).unique()) < maxValues:
                        self.colValueDictTrigger.append(level)

        self.extData = ''
        self.result = ''

        self.levelValues = {}

        self.column = colname

        if pathMeta:
            self.metaData = pd.read_excel(pathMeta)
        else:
            self.metaData = ''

    def resetColWidth(self):
        """ Reset pandas display option for max_colwidth"""
        pd.reset_option('display.max_colwidth')
        return

    def resetSearch(self):
        """Reset self.result to search in full dataframe again."""
        self.result = ''
        return

    def search(self, value):
        """
        Search string in 'colname' column.
        """
        if type(self.result) == str:
            self.result = self.dataframe[self.dataframe[self.column].str.contains(value)]
        elif isinstance(self.result, pd.DataFrame):
            self.result = self.result[self.result[self.column].str.contains(value)]
        return self

    def logicReduce(self, logicList):
        """
        Constructs complex searches for list of (part,value) tuples, connected
        via AND (&),OR (|),NOT (~&) logic. logicList has the format
        '[(part1,value1),&,(part2,value2),|,(part3,value3)]'.
        Evaluation is in order of apperance, e.g. for the above logicList,
        a boolean list is constructed for each (part,value) tuple.
        Then the first two tuples are compared with & yielding a temporary
        result temp1, which is compared with | with the last tuple.
        This creates a resulting boolean list res1,
        which is used to reduce the dataframe.
        """
        self.tempRes = []
        for part in logicList:
            if type(part) == tuple:
                res = self.boolList(part[0], part[1])
                self.tempRes.append(res)
            elif type(part) == str and part in ['&', '|', '~&']:
                self.tempRes.append(part)

        self.res = self.tempRes.pop(0)

        while self.tempRes:
            op = self.tempRes.pop(0)
            res2 = self.tempRes.pop(0)
            if op == '&':
                self.res = self.res & res2
            elif op == '|':
                self.res = self.res | res2
            elif op == '~&':
                self.res = self.res & np.invert(res2)
            else:
                pass

        if type(self.res) != str:
            self.result = self.dataframe[self.res]
        return self

    def boolList(self, level, value):
        """ Returns boolean list for dataframe[part]==value."""
        searchvalue = self._fuzzySearch(level, value)
        if self.dataindex == 'multi' and level != self.column:
            res = self.dataframe.index.get_level_values(level) == self._assertDataType(level, searchvalue, self.dataframe)
        elif self.dataindex == 'single' and level != self.column:
            res = self.dataframe[level] == self._assertDataType(level, searchvalue, self.dataframe)
        elif level == self.column:
            res = self.dataframe[level].str.contains(self._assertDataType(level, value, self.dataframe)) == True
        return res

    def _fuzzySearch(self, level, value):
        """
        Allow fuzzy search for reduce() and search() routines. Limited to
        columns in the colValueDictTrigger list. This excludes columns containing
        numerical data, as well as the main column 'colname'.
        """
        if level in self.colValueDictTrigger and level != self.column:
            if level not in self.levelValues.keys():
                if self.dataindex == 'multi':
                    self.levelValues[level] = self.dataframe.index.get_level_values(level).unique()
                elif self.dataindex == 'single':
                    self.levelValues[level] = self.dataframe[level].unique()
                else:
                    raise ValueError(
                        'DataIndex not "single" or "multi".'
                        )
            else:
                pass

            if value not in self.levelValues[level]:
                try:
                    closestMatch = difflib.get_close_matches(value, self.levelValues[level], 1)
                    if closestMatch:
                        searchValue = closestMatch[0]
                    else:
                        raise ValueError(
                            'Could not find matching expression to search.'
                            )
                except TypeError:
                    searchValue = value
            else:
                searchValue = value
        else:
            searchValue = value
        return searchValue

    def reduce(self, level, value):
        """ Return reduced dataframe for search tuple (level/column,value):
            a) as cross-section for multi-index dataframe:
               df.xs(value,level=level)
            b) as sub-dataframe for single-index dataframe:
               df[df.column == value]

            Result is in self.result, to be able to chain reductions.
            To view result use self.results()
        """

        searchValue = self._fuzzySearch(level, value)
        self._searchString(level, searchValue)
        return self

    def _searchString(self, level, value):
        """Helper function for reducing dataframes"""
        if self.dataindex == 'multi':
            if type(self.result) == str:
                self.result = self.dataframe.xs(self._assertDataType(level, value, self.dataframe), level=level, drop_level=False)
            else:
                self.result = self.result.xs(self._assertDataType(level, value, self.result), level=level, drop_level=False)
        elif self.dataindex == 'single':
            if type(self.result) == str:
                self.result = self.dataframe[self.dataframe[level] == self._assertDataType(level, value, self.dataframe)]
            else:
                self.result = self.result[self.result[level] == self._assertDataType(level, value, self.result)]

    def _assertDataType(self, level, value, dataframe):
        """Helper function to assert correct datatype for value at level or in column"""
        numTypes = ['int8', 'int16', 'int32', 'int64', 'float', 'float64']
        valueType = type(value)
        if self.dataindex == 'multi':
            levelType = dataframe.index.get_level_values(level=level).dtype.name
        elif self.dataindex == 'single':
            levelType = self.dataframe[level].dtype.name
        if levelType == 'object' and valueType == str:
            return value
        elif levelType == 'object' and valueType == int:
            try:
                return str(value)
            except ValueError as err:
                raise('Can not cast {0} to type {1}'.format(value, levelType))
        elif levelType in numTypes and valueType == int:
            return value
        elif levelType in numTypes and valueType == float:
            return value
        elif levelType in numTypes and valueType == str:
            try:
                return int(value)
            except ValueError:
                raise('Can not cast {0} to type {1}'.format(value, levelType))
        else:
            raise ValueError('Can not cast {0} to type {1}'.format(value, levelType))

    def results(self):
        """Returns the search result as a single-index dataframe."""
        if self.dataindex == 'multi':
            self.indexLevels = list(self.result.index.names)
            formatedResult = self.result.reset_index(level=self.indexLevels)
        elif self.dataindex == 'single':
            formatedResult = self.result
        pd.set_option('display.max_colwidth', -1)
        return formatedResult

    def extResults(self, level):
        """
        Returns the search result as a single-index dataframe, extend with
        metadata on the desired level. The metadata is calculated for all other
        levels of multi-indexed dataframe.
        """
        # TODO: Add statistics for single-level dataframes.
        if not self.dataindex == 'multi':
            print('Statistics for multiindexed dataframes only.')
            return
        self.extData = pd.merge(left=self.results(), right=self._metaData(level), on=level)
        cols = [x for x in self.extData.columns if x != self.column]
        cols = cols + [self.column]
        self.extData = self.extData[cols]
        pd.set_option('display.max_colwidth', -1)
        return self.extData

    def _countWords(self, level, value):
        """Helper function to count words on a given level."""
        text = ' '.join(self.dataframe.xs(value, level=level).text.tolist())
        numWords = len(re.findall(r'\w+', text))
        return numWords

    def _getStatistics(self, level):
        """Helper function to return statistics on given level."""
        retDict = {}
        subLevels = self.indexLevels.copy()
        try:
            subLevels.remove(level)
        except ValueError:
            print('{0} not in index'.format(level))

        for part in self.dataframe.index.get_level_values(level).unique():
            partDict = {}
            for subLevel in subLevels:
                num = len(self.dataframe.xs(part, level=level).index.get_level_values(subLevel).unique())
                if num > 1:
                    partDict['Num_' + subLevel] = num
            partDict['words'] = self._countWords(level, part)
            retDict[part] = partDict
        return retDict

    def _metaData(self, level):
        """Helper function to generate dataframe from statistics."""
        resDict = self._getStatistics(level)
        self.metaDataFrame = pd.DataFrame(resDict).fillna('1').astype('int').transpose().reset_index().rename({'index': level}, axis=1)
        return self.metaDataFrame
